@@ -32,8 +32,8 @@ enum {
 };
 
 struct http_request {
-	http_client_callback_t *callback;
-	void *callback_data;
+	const struct http_client_handler *handler;
+	void *handler_ctx;
 
 	/** the CURL easy handle */
 	CURL *curl;
@@ -72,6 +72,12 @@ static struct {
 	bool locked;
 } http_client;
 
+static inline GQuark
+curl_quark(void)
+{
+    return g_quark_from_static_string("curl");
+}
+
 /**
  * Frees all resources of a #http_request object.  Also unregisters
  * the CURL easy handle from the CURL multi handle.  This function
@@ -183,28 +189,30 @@ http_client_update_fds(void)
 
 /**
  * Aborts and frees a running HTTP request and report an error to its
- * callback.
+ * handler.
  */
 static void
-http_request_abort(struct http_request *request)
+http_request_abort(struct http_request *request, GError *error)
 {
 	http_client.requests = g_slist_remove(http_client.requests, request);
 
-	request->callback(0, NULL, request->callback_data);
+	request->handler->error(error, request->handler_ctx);
 	http_request_free(request);
 }
 
 /**
- * Abort and free all HTTP requests, but don't invoke their callback
- * functions.
+ * Abort and free all HTTP requests, but don't invoke their handler
+ * methods.
  */
 static void
-http_client_abort_all_requests(void)
+http_client_abort_all_requests(GError *error)
 {
 	while (http_client.requests != NULL) {
 		struct http_request *request = http_client.requests->data;
-		http_request_abort(request);
+		http_request_abort(request, g_error_copy(error));
 	}
+
+	g_error_free(error);
 }
 
 /**
@@ -228,16 +236,23 @@ http_client_find_request(CURL *curl)
  * A HTTP request is finished: invoke its callback and free it.
  */
 static void
-http_request_done(struct http_request *request, CURLcode result)
+http_request_done(struct http_request *request, CURLcode result, long status)
 {
-	/* invoke the callback function */
-	if (result == CURLE_OK)
-		request->callback(request->body->len, request->body->str,
-				  request->callback_data);
-	else {
-		g_warning("curl failed: %s", request->error);
-		request->callback(0, NULL, request->callback_data);
-	}
+	/* invoke the handler method */
+	if (result != CURLE_OK) {
+		GError *error = g_error_new(curl_quark(), result,
+					    "curl failed: %s",
+					    request->error);
+		request->handler->error(error, request->handler_ctx);
+	} else if (status < 200 || status >= 300) {
+		GError *error = g_error_new(curl_quark(), 0,
+					    "got HTTP status %ld",
+					    status);
+		request->handler->error(error, request->handler_ctx);
+	} else
+		request->handler->response(request->body->len,
+					   request->body->str,
+					   request->handler_ctx);
 
 	/* remove it from the list and free resources */
 	http_client.requests = g_slist_remove(http_client.requests, request);
@@ -263,7 +278,11 @@ http_multi_info_read(void)
 				http_client_find_request(msg->easy_handle);
 			assert(request != NULL);
 
-			http_request_done(request, msg->data.result);
+			long status = 0;
+			curl_easy_getinfo(msg->easy_handle,
+					  CURLINFO_RESPONSE_CODE, &status);
+
+			http_request_done(request, msg->data.result, status);
 		}
 	}
 
@@ -285,9 +304,10 @@ http_multi_perform(void)
 	} while (mcode == CURLM_CALL_MULTI_PERFORM);
 
 	if (mcode != CURLM_OK && mcode != CURLM_CALL_MULTI_PERFORM) {
-		g_warning("curl_multi_perform() failed: %s\n",
-			  curl_multi_strerror(mcode));
-		http_client_abort_all_requests();
+		GError *error = g_error_new(curl_quark(), mcode,
+					    "curl_multi_perform() failed: %s",
+					    curl_multi_strerror(mcode));
+		http_client_abort_all_requests(error);
 		return false;
 	}
 
@@ -442,31 +462,38 @@ http_request_writefunction(void *ptr, size_t size, size_t nmemb, void *stream)
 
 	g_string_append_len(request->body, ptr, size * nmemb);
 
-	if (request->body->len > MAX_RESPONSE_BODY)
+	if (request->body->len > MAX_RESPONSE_BODY) {
 		/* response body too large */
-		http_request_abort(request);
+		GError *error =
+			g_error_new_literal(curl_quark(), 0,
+					    "response body is too large");
+		http_request_abort(request, error);
+	}
 
 	return size * nmemb;
 }
 
 void
 http_client_request(const char *url, const char *post_data,
-		    http_client_callback_t *callback, void *data)
+		    const struct http_client_handler *handler, void *ctx)
 {
 	struct http_request *request = g_new(struct http_request, 1);
 	CURLcode code;
 	CURLMcode mcode;
 	bool success;
 
-	request->callback = callback;
-	request->callback_data = data;
+	request->handler = handler;
+	request->handler_ctx = ctx;
 
 	/* create a CURL request */
 
 	request->curl = curl_easy_init();
 	if (request->curl == NULL) {
 		g_free(request);
-		callback(0, NULL, data);
+
+		GError *error = g_error_new_literal(curl_quark(), 0,
+						    "curl_easy_init() failed");
+		handler->error(error, ctx);
 		return;
 	}
 
@@ -474,7 +501,10 @@ http_client_request(const char *url, const char *post_data,
 	if (mcode != CURLM_OK) {
 		curl_easy_cleanup(request->curl);
 		g_free(request);
-		callback(0, NULL, data);
+
+		GError *error = g_error_new_literal(curl_quark(), 0,
+						    "curl_multi_add_handle() failed");
+		handler->error(error, ctx);
 		return;
 	}
 
@@ -504,7 +534,10 @@ http_client_request(const char *url, const char *post_data,
 		curl_multi_remove_handle(http_client.multi, request->curl);
 		curl_easy_cleanup(request->curl);
 		g_free(request);
-		callback(0, NULL, data);
+
+		GError *error = g_error_new_literal(curl_quark(), code,
+						    "curl_easy_setopt() failed");
+		handler->error(error, ctx);
 		return;
 	}
 
@@ -519,7 +552,10 @@ http_client_request(const char *url, const char *post_data,
 		http_client.requests = g_slist_remove(http_client.requests,
 						      request);
 		http_request_free(request);
-		callback(0, NULL, data);
+
+		GError *error = g_error_new_literal(curl_quark(), code,
+						    "http_multi_perform() failed");
+		handler->error(error, ctx);
 		return;
 	}
 
