@@ -1,5 +1,5 @@
 /* mpdscribble (MPD Client)
- * Copyright (C) 2008-2009 The Music Player Daemon Project
+ * Copyright (C) 2008-2010 The Music Player Daemon Project
  * Copyright (C) 2005-2008 Kuno Woudt <kuno@frob.nl>
  * Project homepage: http://musicpd.org
  
@@ -21,6 +21,7 @@
 #include "journal.h"
 #include "record.h"
 
+#include <assert.h>
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
@@ -28,16 +29,30 @@
 
 static int journal_file_empty;
 
+static void
+journal_write_string(FILE *file, char field, const char *value)
+{
+	if (value != NULL)
+		fprintf(file, "%c = %s\n", field, value);
+}
+
 static void
 journal_write_record(gpointer data, gpointer user_data)
 {
 	struct record *record = data;
 	FILE *file = user_data;
 
+	assert(record->source != NULL);
+
+	journal_write_string(file, 'a', record->artist);
+	journal_write_string(file, 't', record->track);
+	journal_write_string(file, 'b', record->album);
+	journal_write_string(file, 'n', record->number);
+	journal_write_string(file, 'm', record->mbid);
+	journal_write_string(file, 'i', record->time);
+
 	fprintf(file,
-		"a = %s\nt = %s\nb = %s\nm = %s\n"
-		"i = %s\nl = %i\no = %s\n\n", record->artist,
-		record->track, record->album, record->mbid, record->time,
+		"l = %i\no = %s\n\n",
 		record->length, record->source);
 }
 
@@ -176,6 +191,8 @@ void journal_read(const char *path, GQueue *queue)
 			record.track = g_strdup(value);
 		else if (!strcmp("b", key))
 			record.album = g_strdup(value);
+		else if (!strcmp("n", key))
+			record.number = g_strdup(value);
 		else if (!strcmp("m", key))
 			record.mbid = g_strdup(value);
 		else if (!strcmp("i", key))
