@@ -314,7 +314,7 @@ next_line(const char **input_r, const char *end)
 }
 
 static void
-scrobbler_handshake_callback(size_t length, const char *response, void *data)
+scrobbler_handshake_response(size_t length, const char *response, void *data)
 {
 	struct scrobbler *scrobbler = data;
 	const char *end = response + length;
@@ -327,13 +327,6 @@ scrobbler_handshake_callback(size_t length, const char *response, void *data)
 
 	scrobbler->state = SCROBBLER_STATE_NOTHING;
 
-	if (!length) {
-		g_warning("[%s] handshake timed out", scrobbler->config->name);
-		scrobbler_increase_interval(scrobbler);
-		scrobbler_schedule_handshake(scrobbler);
-		return;
-	}
-
 	line = next_line(&response, end);
 	ret = scrobbler_parse_handshake_response(scrobbler, line);
 	g_free(line);
@@ -377,6 +370,30 @@ scrobbler_handshake_callback(size_t length, const char *response, void *data)
 	scrobbler_submit(scrobbler);
 }
 
+static void
+scrobbler_handshake_error(GError *error, void *data)
+{
+	struct scrobbler *scrobbler = data;
+
+	assert(scrobbler != NULL);
+	assert(scrobbler->config->file == NULL);
+	assert(scrobbler->state == SCROBBLER_STATE_HANDSHAKE);
+
+	scrobbler->state = SCROBBLER_STATE_NOTHING;
+
+	g_warning("[%s] handshake error: %s",
+		  scrobbler->config->name, error->message);
+	g_error_free(error);
+
+	scrobbler_increase_interval(scrobbler);
+	scrobbler_schedule_handshake(scrobbler);
+}
+
+static const struct http_client_handler scrobbler_handshake_handler = {
+	.response = scrobbler_handshake_response,
+	.error = scrobbler_handshake_error,
+};
+
 static void
 scrobbler_queue_remove_oldest(GQueue *queue, unsigned count)
 {
@@ -389,7 +406,7 @@ scrobbler_queue_remove_oldest(GQueue *queue, unsigned count)
 }
 
 static void
-scrobbler_submit_callback(size_t length, const char *response, void *data)
+scrobbler_submit_response(size_t length, const char *response, void *data)
 {
 	struct scrobbler *scrobbler = data;
 	char *newline;
@@ -398,14 +415,6 @@ scrobbler_submit_callback(size_t length, const char *response, void *data)
 	assert(scrobbler->state == SCROBBLER_STATE_SUBMITTING);
 	scrobbler->state = SCROBBLER_STATE_READY;
 
-	if (!length) {
-		scrobbler->pending = 0;
-		g_warning("[%s] submit timed out", scrobbler->config->name);
-		scrobbler_increase_interval(scrobbler);
-		scrobbler_schedule_submit(scrobbler);
-		return;
-	}
-
 	newline = memchr(response, '\n', length);
 	if (newline != NULL)
 		length = newline - response;
@@ -442,6 +451,29 @@ scrobbler_submit_callback(size_t length, const char *response, void *data)
 	}
 }
 
+static void
+scrobbler_submit_error(GError *error, void *data)
+{
+	struct scrobbler *scrobbler = data;
+
+	assert(scrobbler->config->file == NULL);
+	assert(scrobbler->state == SCROBBLER_STATE_SUBMITTING);
+
+	scrobbler->state = SCROBBLER_STATE_READY;
+
+	g_warning("[%s] submit error: %s",
+		  scrobbler->config->name, error->message);
+	g_error_free(error);
+
+	scrobbler_increase_interval(scrobbler);
+	scrobbler_schedule_submit(scrobbler);
+}
+
+static const struct http_client_handler scrobbler_submit_handler = {
+	.response = scrobbler_submit_response,
+	.error = scrobbler_submit_error,
+};
+
 char *as_timestamp(void)
 {
 	/* create timestamp for 1.2 protocol. */
@@ -532,7 +564,7 @@ scrobbler_handshake(struct scrobbler *scrobbler)
 	//  notice ("handshake url:\n%s", url);
 
 	http_client_request(url->str, NULL,
-			    &scrobbler_handshake_callback, scrobbler);
+			    &scrobbler_handshake_handler, scrobbler);
 
 	g_string_free(url, true);
 }
@@ -594,7 +626,7 @@ scrobbler_send_now_playing(struct scrobbler *scrobbler, const char *artist,
 
 	http_client_request(scrobbler->nowplay_url,
 			    post_data->str,
-			    scrobbler_submit_callback, scrobbler);
+			    &scrobbler_submit_handler, scrobbler);
 
 	g_string_free(post_data, true);
 }
@@ -700,7 +732,7 @@ scrobbler_submit(struct scrobbler *scrobbler)
 	scrobbler->pending = count;
 	http_client_request(scrobbler->submit_url,
 			    post_data->str,
-			    &scrobbler_submit_callback, scrobbler);
+			    &scrobbler_submit_handler, scrobbler);
 
 	g_string_free(post_data, true);
 }
